{-
Name: Jeeho Song 
Uni: js4892 

COMS 4995 003 Parallel Function Programming
Final Project Code 

Reference: 
https://hackage.haskell.org (refer. for haskell)
https://wiki.haskell.org (refer. for haskell)
https://github.com/gregorulm/h2048/blob/master/h2048.hs (refer. for 2048 Game base code)

compile and execute : 
$ ghc -threaded -rtsopts -eventlog --make 2048.hs 
$ time ./2048 +RTS -ls -s
$ time ./2048 +RTS -N2 -ls -s
$ time ./2048 +RTS -N4 -ls -s
-}
import Prelude hiding (Left, Right)
import Data.List
import Data.Maybe
import System.IO
import System.Random -- cabal install --lib random (to install random)
import Text.Printf
import Control.Parallel.Strategies
import InteractiveEval (Term(val))

{-
Parallel Implementation with using Strategies 
"Strategies" provide methods for parallel implementation. 
to install import Strategies : 
$ cabal install --lib parallel 
import Control.Parallel.Strategies
-}

-- type for 4 options that players can take in each turn 
data Move = Up | Down | Left | Right
-- type grid as list of list of int 
type Grid = [[Int]]
maxVal = toInteger (maxBound :: Int)
minVal = toInteger (minBound :: Int)
direction = [Left, Right, Up, Down]
start_score = 0 

-- <defining functions>
-- start : initialize a 4x4 grid with 0 value
start :: IO Grid
start = do grid' <- addRandomTile $ replicate 4 [0, 0, 0, 0]
           addRandomTile grid'

-- addRandomTile : add a value of 2 (90%) or 4(10%) in a tile with 0 value 
addRandomTile :: Grid -> IO Grid
addRandomTile grid = do
    val <- pickRandomEmptySpot [2,2,2,2,2,2,2,2,2,4]
    let randomSpot = getCoordinateOfZero grid
    selected_random_spot <- pickRandomEmptySpot randomSpot
    let new_grid = setNewGrid grid selected_random_spot val
    return new_grid

-- getCoordinateOfZero : get coordinates of tiles with 0 value 
getCoordinateOfZero :: Grid -> [(Int, Int)]
getCoordinateOfZero grid = filter (\(row, col) -> (grid!!row)!!col == 0) coordinates
    where singleRow n = zip (replicate 4 n) [0..3]
          coordinates = concatMap singleRow [0..3]

-- pickRandomEmptySpot : choose randomly where a new tile with 2 or 4 is set among an empty tile
pickRandomEmptySpot :: [a] -> IO a
pickRandomEmptySpot xs = do
    i <- randomRIO (0, length xs-1)
    return (xs !! i)

-- setNewGrid : set a new grid configuration 
setNewGrid :: Grid -> (Int, Int) -> Int -> Grid
setNewGrid grid (row, col) val = pre ++ [mid] ++ post
    where pre  = take row grid
          mid  = take col (grid!!row) ++ [val] ++ drop (col + 1) (grid!!row)
          post = drop (row + 1) grid

-- merge : combine two same numbered tiles 
merge :: [Int] -> [Int]
merge xs = merged ++ padding
    where padding = replicate (length xs - length merged) 0
          merged  = combine $ filter (/= 0) xs
          combine (x:y:ys) | x == y = x * 2 : combine ys
                           | otherwise = x  : combine (y:ys)
          combine x = x

-- move : get a new grid after move with direction (up, down, left, right)
move :: Move -> Grid -> Grid
move Left  = map merge
move Right = map (reverse . merge . reverse)
move Up    = transpose . move Left  . transpose
move Down  = transpose . move Right . transpose

-- isMoveLeft : check if it is applied in "LEFT"
isMoveLeft :: Grid -> Bool
isMoveLeft grid = sum allChoices > 0
    where allChoices = map (length . getCoordinateOfZero . flip move grid) directions
          directions = [Left, Right, Up, Down]

-- printGrid : display the current grid 
printGrid :: Grid -> IO ()
printGrid grid = do
    putStr "\n"
    mapM_ (putStrLn . concatMap (printf "%5d")) grid

-- checkGoalSate : check if it is a goal state (a tile with the value of 2048)
checkGoalSate :: Grid -> Bool
checkGoalSate grid = [] /= filter (== 2048) (concat grid)

-- get child nodes grid array 
getChildNodes :: Grid -> [Grid]
getChildNodes grid = filter isNotGrid [move direction grid | direction <- direction ]
    where isNotGrid values = values /= grid

-- goal state incentive + 3 hueristic methods 
-- getWeightedValue : get weighted value (after applying heuristic methods)
getWeightedValue :: Grid -> Integer
getWeightedValue grid = maximum [if x == 2048 then 1000 else 0 | x <- map maximum grid] + applyPositionWeights grid - applySimilarityWeights grid + (100*getSpotZero grid)

-- 3 Heuristic Methods 
getSpotZero :: Grid -> Integer
getSpotZero grid = toInteger $ sum $ map checkZeroSpot grid
    where checkZeroSpot grid = length $ filter isZero grid
            where isZero x = x == 0

-- applyPositionWeights : new values of grid in cosideration of position heuristically  
applyPositionWeights :: Grid -> Integer
applyPositionWeights grid = toInteger $ sum (map sum (zipWith (zipWith (*)) weights grid))
    where weights = [[10,8,6,4],[8,8,6,2],[6,6,4,1],[4,2,1,1]]

-- applySimilarityWeights : get tiles with similar values closer 
applySimilarityWeights :: Grid -> Integer
applySimilarityWeights [] = 0
applySimilarityWeights (x:xs) = smc x + applySimilarityWeights xs
    where smc (a:b:c:d:_) = fromIntegral $ 10*(abs(a-b) + abs(b-c) + abs(c-d))
            where smc [] = 0


-- parallel implementation for minimax (using Strategies)
-- getMaximumGrid : find a next grid that maximizes the winning chance in paralell implementation 
getMaximumGrid :: Grid -> Int -> Grid
getMaximumGrid grid max_depth = new_grid !! fromJust (elemIndex (maximum values) values)
    where new_grid = getChildNodes grid          
          values = runEval $ do {
              ; result <- rpar [getDepthValue grid (max_depth-1) | grid <- new_grid]              
              ; rseq result
            --   ; rdeepseq result
              ; return result
          }
    -- runEval :: Eval a -> a (to pull the result out of the monad)
    -- parList :: Strategy a -> Strategy [a] (to evaluate each element of a list in parallel according to given strategy)
    -- rpar :: Strategy a (to spark its argument (for evaluation in parallel))
    -- rseq :: Strategy a (to evaluates its argument to weak head normal form)
    -- rdeepseq :: NFData a => Strategy a (to fully evaluates its argument)

-- getDepthValue : get values of nodes in the depth  
getDepthValue :: Grid -> Int -> Integer
getDepthValue grid 0 = getWeightedValue grid
getDepthValue grid depth = toInteger twoVal + toInteger fourVal
    where twoVal = round(realToFrac(minimizeVal grid (getCoordinateOfZero grid) 2 minVal maxVal maxVal (depth-1))*0.9)
          fourVal = round(realToFrac(minimizeVal grid (getCoordinateOfZero grid) 4 minVal maxVal maxVal (depth-1))*0.9)

-- minimax algorithm : minimizeVal & maxmizeVal 
minimizeVal :: Grid -> [(Int, Int)] -> Int -> Integer -> Integer -> Integer -> Int -> Integer
minimizeVal grid [] value alpha beta min depth = min
minimizeVal grid (x:xs) value alpha beta min depth
  | depth == 0 = getWeightedValue grid
  | val < min = minimizeVal grid (x:xs) value alpha beta val depth
  | beta > min = minimizeVal grid (x:xs) value alpha min min depth
  | alpha >= min = minimizeVal grid xs value alpha beta min depth
  | otherwise = minimizeVal grid xs value alpha beta min depth
    where (netSet, val) = maximizeVal (getChildNodes (setNewGrid grid x value)) grid alpha beta minVal (depth-1)

maximizeVal :: [Grid] -> Grid -> Integer -> Integer -> Integer -> Int -> (Grid, Integer)
maximizeVal [] grid alpha beta max max_depth = (grid, max)
maximizeVal (x:xs) grid alpha beta max max_depth
  | max_depth == 0 = (grid, getWeightedValue grid)
  | val > max = maximizeVal (x:xs) x alpha beta val max_depth
  | max >= beta = maximizeVal xs grid alpha beta max max_depth
  | max > alpha = maximizeVal (x:xs) grid max beta max max_depth
  | otherwise = maximizeVal xs grid alpha beta max max_depth
    where val = getDepthValue x (max_depth-1)

-- gameLoop : implement loop of game 
gameLoop :: Grid -> Int -> IO ()
gameLoop grid num_of_move
    | isMoveLeft grid = do
        printGrid grid
        if checkGoalSate grid
        then do putStrLn "You win"
                print num_of_move
        else do let new_grid = getMaximumGrid grid 6
                if grid /= new_grid
                then do new <- addRandomTile new_grid
                        gameLoop new (num_of_move+1)
                else gameLoop grid (num_of_move+1)
    | otherwise = do
        printGrid grid
        putStrLn "You lose"
        print num_of_move 

main :: IO ()
main = do
    hSetBuffering stdin NoBuffering
    grid <- start
    gameLoop grid start_score
